import json


class Project:
    """
    Model an Ingest Project entity
    """

    def __init__(self, project_id, ingest_api_agent):
        self.project_id = project_id
        self.api = ingest_api_agent
        self.data = None
        self._load()

    @property
    def uuid(self):
        return self.data['uuid']

    def submission_envelopes(self):
        data = self.api.get(self.data['_links']['submissionEnvelopes']['href'])
        return [
            SubmissionEnvelope(data=subm_data, ingest_api_agent=self.api)
            for subm_data in data['_embedded']['submissionEnvelopes']
        ]

    def _load(self):
        self.data = self.api.get(f"/projects/{self.project_id}")


class SubmissionEnvelope:
    """
    Model an Ingest Submission Envelope entity
    """

    # May be primed wih data, or of you supply an ID, we will go get the data

    def __init__(self, ingest_api_agent, envelope_id=None, data=None):
        if not envelope_id and not data:
            raise RuntimeError("either envelope_id or data must be provided")
        self.api = ingest_api_agent
        self.data = None
        if envelope_id:
            self.envelope_id = envelope_id
            self._load()
        else:
            self.data = data
            self.envelope_id = data['_links']['self']['href'].split('/')[-1]

    def __str__(self):
        return f"SubmissionEnvelope\n\tid={self.envelope_id}\n\tuuid={self.uuid}\n\tstatus={self.status}"

    def show_associated(self, entities_to_show, verbose=False):
        self.verbose = verbose
        if 'bundles' in entities_to_show:
            if self.status == 'Complete':
                print("\tBundles:")
                for bundle in self.bundles():
                    print(f"\t\t{bundle}")

        if 'files' in entities_to_show:
                print("\tFiles:")
                for file in self.files():
                    self._output(file)

    def _output(self, thing):
        if self.verbose:
            print("%r" % thing)
        else:
            print(thing)

    def files(self):
        return [File(file_data) for file_data in
                self.api.get_all(self.data['_links']['files']['href'], 'files')]

    def iter_files(self):
        url = self.data['_links']['files']['href']
        for page in self.api.iter_pages(url):
            for file in page['files']:
                yield file

    def reload(self):
        self._load()
        return self

    @property
    def status(self):
        return self.data['submissionState']

    @property
    def uuid(self):
        return self.data['uuid']['uuid']

    def upload_credentials(self):
        """ Return upload area credentials or None if this envelope doesn't have an upload area yet """
        staging_details = self.data.get('stagingDetails', None)
        if staging_details and 'stagingAreaLocation' in staging_details:
            return staging_details.get('stagingAreaLocation', {}).get('value', None)
        return None

    def bundles(self):
        url = self.data['_links']['bundleManifests']['href']
        manifests = self.api.get_all(url, 'bundleManifests')
        return [manifest['bundleUuid'] for manifest in manifests]

    def _load(self):
        self.data = self.api.get(f"/submissionEnvelopes/{self.envelope_id}")


class File:
    """
    Model an Ingest File entity
    """

    def __init__(self, file_data):
        self._data = file_data

    def __str__(self):
        return (f"\t\tfileName {self.name}\n" +
                f"\t\tcloudUrl {self.cloud_url}\n")

    def __repr__(self):
        return json.dumps(self._data, indent=2)

    @property
    def name(self):
        return self._data['fileName']

    @property
    def cloud_url(self):
        return self._data['cloudUrl']



import json
import os

import requests


class DataStoreAgent:

    DSS_API_URL_TEMPLATE = "https://dss.{deployment}.data.humancellatlas.org/v1"

    def __init__(self, deployment):
        self.deployment = deployment
        self.dss_url = self.DSS_API_URL_TEMPLATE.format(deployment=deployment)

    def search(self, query, replica='aws'):
        url = f"{self.dss_url}/search?replica={replica}"
        response = requests.post(url, json={"es_query": query})
        return response.json()['results']

    def download_bundle(self, bundle_uuid, target_folder):
        print(f"Downloading bundle {bundle_uuid}:")
        manifest = self.bundle_manifest(bundle_uuid)
        bundle_folder = os.path.join(target_folder, bundle_uuid)
        try:
            os.makedirs(bundle_folder)
        except FileExistsError:
            pass

        for f in manifest['bundle']['files']:
            self.download_file(f['uuid'], save_as=os.path.join(bundle_folder, f['name']))
        return bundle_folder

    def bundle_manifest(self, bundle_uuid, replica='aws'):
        url = f"{self.dss_url}/bundles/{bundle_uuid}?replica={replica}"
        response = requests.get(url)
        assert response.ok
        assert response.headers['Content-type'] == 'application/json'
        return json.loads(response.content)

    def download_file(self, file_uuid, save_as, replica='aws'):
        url = f"{self.dss_url}/files/{file_uuid}?replica={replica}"
        print(f"Downloading file {file_uuid} to {save_as}")
        response = requests.get(url, stream=True)
        assert response.ok
        with open(save_as, 'wb') as f:
            for chunk in response.iter_content(chunk_size=1024):
                if chunk:  # filter out keep-alive new chunks
                    f.write(chunk)



from urllib.parse import urlencode

import requests

from .ingest_auth_agent import IngestAuthAgent
from .ingest_entities import Project, SubmissionEnvelope


class IngestApiAgent:

    def __init__(self, deployment):
        self.deployment = deployment
        self.ingest_api_url = self._ingest_api_url()
        self.auth_headers = IngestAuthAgent().make_auth_header()

    def project(self, project_id):
        return Project(project_id=project_id, ingest_api_agent=self)

    def submission(self, submission_id):
        return SubmissionEnvelope(envelope_id=submission_id, ingest_api_agent=self)

    def iter_submissions(self):
        for page in self.iter_pages('/submissionEnvelopes', page_size=500, sort='submissionDate,desc'):
            for submission_data in page['submissionEnvelopes']:
                yield SubmissionEnvelope(data=submission_data, ingest_api_agent=self)

    """
    Get a collection resource.
    Iterates through all pages gathering results and returns a list.
    """
    def get_all(self, path_or_url, result_element_we_are_interested_in):
        results = []
        for page in self.iter_pages(path_or_url):
            results += page[result_element_we_are_interested_in]
        return results

    """
    Iterate through a collection using HATEOAS pagination, yielding pages.
    """
    def iter_pages(self, path_or_url, page_size=100, sort=None):
        url_params = {'size': page_size}
        if sort:
            url_params['sort'] = sort
        path_or_url += '?' + urlencode(url_params)

        while True:
            data = self.get(path_or_url)
            if '_embedded' not in data:
                break

            yield data['_embedded']

            if 'next' in data['_links']:
                path_or_url = data['_links']['next']['href']
            else:
                break

    """
    Get a singleton resource.
    """
    def get(self, path_or_url):
        if path_or_url.startswith('http'):
            url = path_or_url
        else:
            url = f"{self.ingest_api_url}{path_or_url}"

        response = requests.get(url, headers=self.auth_headers)

        if response.ok:
            return response.json()
        else:
            raise RuntimeError(f"GET {url} got {response}")

    def _ingest_api_url(self):
        if self.deployment == 'prod':
            return "http://api.ingest.data.humancellatlas.org"
        else:
            return f"http://api.ingest.{self.deployment}.data.humancellatlas.org"
